import hashlib
import multiprocessing as mp
import signal
import subprocess
import sys
import time
import types
from datetime import date,timedelta,datetime
from enum import Enum
from functools import partial
from pprint import pformat
from typing import Iterable,Sized

import dateutil.parser
import numpy as np
import pandas as pd
import ruamel.yaml as raml
from sklearn.preprocessing import MinMaxScaler

from . import log_support as logsup
from .helpers import constrained_sequence
from .locations import *

# pd.set_option('max_colwidth',40)
# pd.set_option('expand_frame_repr',False)
pd.set_option('display.max_colwidth',-1)

def func_has_classarg(func,args):
	has_classarg=False
	if args:
		funcattr=getattr(args[0],func.__name__,None)
		if funcattr is not None and hasattr(funcattr,'__self__'):
			has_classarg=True
	return has_classarg

def reprer(obj):
	if isinstance(obj,dict): d=obj
	else: d=obj.__dict__
	body=[f"{k:<20}: {('None' if v is None else v)}" for k,v in d.items() if k[0]!='_']
	if body: delim="\n"
	else: body,delim=["Expired!"]," "
	return f'"{obj.__class__.__name__}" item:'+delim+"\n".join(body)

class AutoName(Enum):
	
	def _generate_next_value_(name,start,count,last_values):
		return name

def run_shell(*args,non_blocking=False,chdir=None,with_popen=False,ret_triggers=None,**kwargs):
	chdir=chdir or os.getcwd()
	normal_dir=os.getcwd()
	os.chdir(chdir)
	
	if non_blocking:
		subprocess.Popen(args,stdout=subprocess.PIPE,**kwargs)
	else:
		#output=subprocess.run(args,stdout=subprocess.PIPE,**kwargs).stdout.decode()
		#output=subprocess.run(args,stdout=subprocess.PIPE,**kwargs).stdout.decode()
		if with_popen:
			process=subprocess.Popen(args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,preexec_fn=os.setsid)
			regular_termination=True
			nextline=''
			while regular_termination:
				nextline=process.stdout.readline().decode()
				# print('nextline:',nextline)
				if nextline=='' and process.poll() is not None:
					break
				if ret_triggers:
					for ret_trigger in ret_triggers:
						if ret_trigger in nextline:
							regular_termination=False
							break
				logsup.logger.debug(nextline)
				# sys.stdout.write(nextline)
				sys.stdout.flush()
			if regular_termination:
				output=process.communicate()[0]
				exitCode=process.returncode
				if exitCode!=0:
					raise Exception(args,exitCode,output)
			else:
				logsup.logger.info(r'Manual process termination triggered with line:\n%s',nextline)
				logsup.logger.debug('Killing process: %s',process.pid)
				os.killpg(os.getpgid(process.pid),signal.SIGTERM)
				raise Exception('Manual process termination.')
		else:
			subprocess.run(args,stdout=subprocess.PIPE,**kwargs).stdout.decode()
	os.chdir(normal_dir)

def period_within(days=None,start_date=None,end_date=None):
	if isinstance(start_date,str): start_date=dateutil.parser.parse(start_date)
	if isinstance(end_date,str): end_date=dateutil.parser.parse(end_date)
	if days is not None:
		if start_date is not None:
			end_date=start_date+timedelta(days=days)
		elif end_date is not None:
			start_date=end_date-timedelta(days=days)
		else:
			end_date=date.today()
			start_date=end_date-timedelta(days=days)
	else:
		end_date=end_date or date.today()
	return start_date,end_date

def datetime_within(start=None,end=None,**timedelta_opts):
	if timedelta_opts:
		if start is not None:
			end=start+timedelta(**timedelta_opts)
		elif end is not None:
			start=end-timedelta(**timedelta_opts)
		else:
			end=datetime.now()
			start=end-timedelta(**timedelta_opts)
	else:
		end=end or datetime.now()
	return start,end

def try_to_do(func,*args,check_success=None,max_attempts=5,max_wait_seconds=0,**kwargs):
	def success_checker(x):
		assert x is not None
		return x
	
	if check_success is None: check_success=success_checker
	max_attempts=max(max_attempts,1)
	if max_wait_seconds:
		waiting_times=constrained_sequence(max_attempts,lambda x:np.exp(x/4))
		waiting_times=MinMaxScaler((0,max_wait_seconds))\
			.fit_transform(np.array(waiting_times).reshape(-1,1))
	else: waiting_times=[0]*max_attempts
	logsup.logger.debug('Trying to accomplish "%s" ...',func.__name__)
	wait_time=iter(waiting_times)
	for i in range(max_attempts):
		time_to_wait=next(wait_time)
		if time_to_wait!=0:
			logsup.logger.info('Sleeping for %s sec.',time_to_wait)
			time.sleep(time_to_wait)
		logsup.logger.debug('Attempt #%s.',i+1)
		out=func(*args,**kwargs)
		# simple_logger.debug('check_success:')
		if check_success(out):
			logsup.logger.debug('Succeeded.')
			return out
		else:
			logsup.logger.info('Attempt #%s failed.',i+1)
	return False

def run_until_victory(func,retry_timeout,*args,**kwargs):
	manager=mp.Manager()
	res_dict=manager.dict()
	args=[res_dict,*args]
	count=0
	while not res_dict:
		count+=1
		logsup.logger.info(f'Starting "{func.__name__}". Counter: {count}')
		p=mp.Process(target=func,args=args,kwargs=kwargs)
		p.start()
		p.join(retry_timeout)
		p.terminate()
	return res_dict

def run_processes_with_map_async(proc_func,payload: list,proc_count=None):
	proc_count=proc_count or (mp.cpu_count()-1)
	with mp.Pool(proc_count) as p:
		results=list()
		p.map_async(proc_func,payload,callback=lambda x:results.extend(x))
		p.close()
		p.join()
	return results

# def run_reactive_worker(listening_func,job_func,output_func=None,worker_count=1,interval=10):
# 	with concurrent.futures.ProcessPoolExecutor(worker_count) as executor:
# 		rx.Observable.interval(interval).from_iterable(listening_func()).flat_map(
# 			lambda *args,**kwargs:executor.submit(job_func,*args,**kwargs)
# 			).subscribe(output_func)

# class ReactiveTask(ABC):
#
# 	def __init__(self,filename,interval=10):
# 		self.filename=filename
# 		self.interval=interval
#
# 	def sleep(self):
# 		time.sleep(self.interval)
#
# 	# def create_log_file(self,info_opts=None,**kwargs):
# 	# 	logsup.logger.create_log_file(self.filename,dict(process=self.__class__.__name__,**(info_opts or {})),**kwargs)
#
# 	def data_generator(self):
# 		raise NotImplementedError
#
# 	def data_processor(self,data):
# 		raise NotImplementedError
#
# 	def post_processor(self,output_data):
# 		pass
#
# 	def run(self,worker_count=1):
# 		with concurrent.futures.ThreadPoolExecutor(worker_count) as executor:
# 			# timing_observable=rx.Observable.interval(int(self.interval*1e3))
# 			# timing_observable.from_(self.data_generator()).flat_map(lambda data:executor.submit(run_processes_with_map_async,self.data_processor,payload=[data],proc_count=1))\
# 			# 	.subscribe(self.post_processor)
# 			rx.Observable.from_iterable(iter(self.data_generator()))\
# 				.flat_map(lambda data:executor.submit(
# 				run_processes_with_map_async,self.data_processor,payload=[data],proc_count=1))\
# 				.subscribe(self.post_processor)
# 			# rx.Observable\
# 			# 	.zip(timing_observable,processor_observable,lambda t,output_data:output_data)\
# 			# 	.flat_map(lambda data:executor.submit(run_processes_with_map_async,self.data_processor,payload=[data],proc_count=1))\
# 			# 	.subscribe(self.post_processor)
# 			input('The run finished. Give any input: ')

def infiterate(iter_obj,max_iter_count=None,next_getter=None,inform_count=8,on_inform=None):
	if on_inform is None: on_inform=lambda *args,**kwargs:True
	if isinstance(iter_obj,int):
		an_iterable,iter_count=range(iter_obj),iter_obj
	else:
		an_iterable=iter_obj
		if isinstance(iter_obj,Sized):
			iter_count=len(iter_obj)
		else:
			assert max_iter_count
			iter_count=max_iter_count
	max_iter_count=max_iter_count or iter_count
	if next_getter is None: next_getter=lambda _obj:_obj
	for i,obj in enumerate(an_iterable):
		if i%(max_iter_count//inform_count)==0:
			on_inform()
			logsup.logger.info('%d%%, i=%d',i/(max_iter_count-1)*100,i)
		
		obj=next_getter(obj)
		yield obj
		if i==max_iter_count-1:
			on_inform()
			logsup.logger.info('%d%%, i=%d',i/(max_iter_count-1)*100,i)
			return

class ParallelTasker:
	
	def __init__(self,func,*args,**kwargs):
		self.partial_func=partial(func,*args,**kwargs)
		self.queue=mp.Queue()
		self._total_proc_count=0
		pass
	
	def set_run_params(self,**kwargs):
		val_lengths={len(v) for v in kwargs.values()}
		assert len(val_lengths)==1
		val_length=val_lengths.pop()
		
		for i in range(val_length):
			self.queue.put((i,{k:v[i] for k,v in kwargs.items()}))
			self._total_proc_count+=1
			logsup.logger.info('self._total_proc_count: %s',self._total_proc_count)
		
		return self
	
	def __process_func(self,queue,common_list):
		i,kwargs=queue.get()
		result=self.partial_func(**kwargs)
		common_list.append([i,result])
	
	@staticmethod
	def __join_finished_processes(active_procs,sleep_time):
		while True:
			# simple_logger.info('Process queue is full. Searching for finished processes.')
			for p in active_procs:
				if not p.is_alive():
					logsup.logger.info('Finished process found. Joining it.')
					active_procs.remove(p)
					p.join()
					# p.terminate()
					logsup.logger.info('Process successfully removed.')
					return
			time.sleep(sleep_time)
	
	def run(self,workers_count=None,sleep_time=1.0):
		workers_count=min(mp.cpu_count()-1,workers_count or self._total_proc_count)
		manager=mp.Manager()
		common_list=manager.list()
		processes=[mp.Process(target=self.__process_func,args=(self.queue,common_list)) for _ in range(self.queue.qsize())]
		active_procs=list()
		while True:
			logsup.logger.info('loop begins')
			if len(processes)==0:
				logsup.logger.info('Waiting for the last jobs to finish.')
				for active_p in active_procs:
					active_p.join()
				break
			if len(active_procs)<workers_count:
				logsup.logger.info('Adding a process')
				proc=processes.pop()
				proc.start()
				active_procs.append(proc)
			else:
				self.__join_finished_processes(active_procs,sleep_time)
			time.sleep(sleep_time)
		return [item[1] for item in sorted(common_list,key=lambda x:x[0])]

def try_wrapper(max_tries=12,max_sleep_time=600,pass_error=True,on_pass_return=None):
	# if pass_error: assert on_pass_return is not None
	if max_tries in [None,'inf']: max_tries=float('inf')
	
	def decorator(func):
		def wrapper(*args,**kwargs):
			try_count=1
			while True:
				try:
					out=func(*args,**kwargs)
					if try_count!=1:
						logsup.logger.info('Succeeded at try #%d',try_count)
					break
				except Exception as e:
					if max_tries and max_tries<=try_count:
						logsup.logger.info('Reached max. number of tries (%d).',try_count)
						if pass_error: return on_pass_return
						else: raise
					try_count+=1
					if max_tries==float('inf'): sleep_time=max_sleep_time
					else: sleep_time=max_sleep_time/(1+np.exp(-(try_count-1-max_tries/2)))
					
					logsup.logger.error('error: %s',str(e),exc_info=True)
					logsup.logger.info('Sleeping for %.3f seconds.',sleep_time)
					time.sleep(sleep_time)
					logsup.logger.info('Try #%d',try_count)
			
			return out
		
		return wrapper
	
	return decorator

_try_wrapper_attrs="_obj _try_wrapper_args _try_wrapper_kwargs".split(' ')

class TryWrapper:
	
	def __init__(self,obj,*try_wrapper_args,**try_wrapper_kwargs):
		self._obj=obj
		self._try_wrapper_args=try_wrapper_args
		self._try_wrapper_kwargs=try_wrapper_kwargs
	
	def _func_wrapper(self,func):
		return func
	
	def __getattr__(self,item):
		if item in _try_wrapper_attrs: return self.__dict__[item]
		else:
			attr=getattr(self._obj,item)
			if type(attr) is types.MethodType:
				return try_wrapper(*self._try_wrapper_args,**self._try_wrapper_kwargs)(self._func_wrapper(attr))
			else: return attr

class GenericObjContainer:
	
	def __init__(self,obj):
		self.obj=obj

class ReactiveObj(GenericObjContainer):
	
	def apply(self,func):
		if isinstance(self.obj,Iterable):
			return list(map(func,self.obj))
		else: return func(self.obj)

class Mapper(GenericObjContainer):
	
	def map(self,func):
		return list(map(func,self.obj))

class AttributeItemGetterSetter:
	
	def __init__(self,obj):
		self.obj=obj
	
	def __getattr__(self,item):
		return AttributeItemGetterSetter(self.obj[item])
	
	def __getitem__(self,item):
		return AttributeItemGetterSetter(self.obj[item])
	
	def __setitem__(self,key,value):
		self.obj[key]=value
	
	def __repr__(self):
		return f"AttributeItemGetterSetter({pformat(self.obj)})"
	
	def __call__(self,*args,**kwargs):
		return self.obj

class AttributeItemSetter:
	
	def __init__(self,obj):
		self.obj=obj
	
	def __setattr__(self,key,value):
		self.obj[key]=value
	
	def __setitem__(self,key,value):
		self.obj[key]=value
	
	def __repr__(self):
		return f"AttributeItemSetter({pformat(self.obj)})"

class ConfigReader(AttributeItemGetterSetter):
	
	def __init__(self,config_path):
		self.config_path=config_path
		self.yaml=raml.YAML()
		self.yaml.default_flow_style=False
		super(ConfigReader,self).__init__(self.__load())
	
	def __load(self):
		return self.yaml.load(open(self.config_path))
	
	def update(self):
		self.yaml.dump(self.obj,open(self.config_path,'w'))
		return self.__load()

if __name__=='__main__':
	pass

def md5(filepath):
	hash_md5=hashlib.md5()
	with open(filepath,"rb") as f:
		for chunk in iter(lambda:f.read(4096),b""):
			hash_md5.update(chunk)
	return hash_md5.hexdigest()



import posixpath
import stat
import os
from abc import abstractmethod

from fabric import Connection
from fabric.transfer import Transfer,Result

import lgblkb_tools.utils
from . import global_support as gsup,log_support as logsup

class FabricTransfer(Transfer):
	
	def get(self,remote,local=None,preserve_mode=True):
		"""
		Download a file from the current connection to the local filesystem.

		:param str remote:
			Remote file to download.

			May be absolute, or relative to the remote working directory.

			.. note::
				Most SFTP servers set the remote working directory to the
				connecting user's home directory, and (unlike most shells) do
				*not* expand tildes (``~``).

				For example, instead of saying ``get("~/tmp/archive.tgz")``,
				say ``get("tmp/archive.tgz")``.

		:param local:
			Local path to store downloaded file in, or a file-like object.

			**If None or another 'falsey'/empty value is given** (the default),
			the remote file is downloaded to the current working directory (as
			seen by `os.getcwd`) using its remote filename.

			**If a string is given**, it should be a path to a local directory
			or file and is subject to similar behavior as that seen by common
			Unix utilities or OpenSSH's ``sftp`` or ``scp`` tools.

			For example, if the local path is a directory, the remote path's
			base filename will be added onto it (so ``get('foo/bar/file.txt',
			'/tmp/')`` would result in creation or overwriting of
			``/tmp/file.txt``).

			.. note::
				When dealing with nonexistent file paths, normal Python file
				handling concerns come into play - for example, a ``local``
				path containing non-leaf directories which do not exist, will
				typically result in an `OSError`.

			**If a file-like object is given**, the contents of the remote file
			are simply written into it.

		:param bool preserve_mode:
			Whether to `os.chmod` the local file so it matches the remote
			file's mode (default: ``True``).

		:returns: A `.Result` object.

		.. versionadded:: 2.0
		"""
		# TODO: how does this API change if we want to implement
		# remote-to-remote file transfer? (Is that even realistic?)
		# TODO: handle v1's string interpolation bits, especially the default
		# one, or at least think about how that would work re: split between
		# single and multiple server targets.
		# TODO: callback support
		# TODO: how best to allow changing the behavior/semantics of
		# remote/local (e.g. users might want 'safer' behavior that complains
		# instead of overwriting existing files) - this likely ties into the
		# "how to handle recursive/rsync" and "how to handle scp" questions
		
		# Massage remote path
		if not remote:
			raise ValueError("Remote path must not be empty!")
		orig_remote=remote
		remote=posixpath.join(
			self.sftp.getcwd() or self.sftp.normalize("."),remote
			)
		
		# Massage local path:
		# - handle file-ness
		# - if path, fill with remote name if empty, & make absolute
		orig_local=local
		is_file_like=hasattr(local,"write") and callable(local.write)
		if not local:
			local=posixpath.basename(remote)
		if not is_file_like:
			local=os.path.abspath(local)
		
		# Run Paramiko-level .get() (side-effects only. womp.)
		# TODO: push some of the path handling into Paramiko; it should be
		# responsible for dealing with path cleaning etc.
		# TODO: probably preserve warning message from v1 when overwriting
		# existing files. Use logging for that obviously.
		#
		# If local appears to be a file-like object, use sftp.getfo, not get
		if is_file_like:
			self.sftp.getfo(remotepath=remote,fl=local,confirm=False)
		else:
			self.sftp.get(remotepath=remote,localpath=local)
			# Set mode to same as remote end
			# TODO: Push this down into SFTPClient sometime (requires backwards
			# incompat release.)
			if preserve_mode:
				remote_mode=self.sftp.stat(remote).st_mode
				mode=stat.S_IMODE(remote_mode)
				os.chmod(local,mode)
		# Return something useful
		return Result(
			orig_remote=orig_remote,
			remote=remote,
			orig_local=orig_local,
			local=local,
			connection=self.connection,
			)
	
	def put(self,local,remote=None,preserve_mode=True):
		"""
		Upload a file from the local filesystem to the current connection.

		:param local:
			Local path of file to upload, or a file-like object.

			**If a string is given**, it should be a path to a local (regular)
			file (not a directory).

			.. note::
				When dealing with nonexistent file paths, normal Python file
				handling concerns come into play - for example, trying to
				upload a nonexistent ``local`` path will typically result in an
				`OSError`.

			**If a file-like object is given**, its contents are written to the
			remote file path.

		:param str remote:
			Remote path to which the local file will be written.

			.. note::
				Most SFTP servers set the remote working directory to the
				connecting user's home directory, and (unlike most shells) do
				*not* expand tildes (``~``).

				For example, instead of saying ``put("archive.tgz",
				"~/tmp/")``, say ``put("archive.tgz", "tmp/")``.

				In addition, this means that 'falsey'/empty values (such as the
				default value, ``None``) are allowed and result in uploading to
				the remote home directory.

			.. note::
				When ``local`` is a file-like object, ``remote`` is required
				and must refer to a valid file path (not a directory).

		:param bool preserve_mode:
			Whether to ``chmod`` the remote file so it matches the local file's
			mode (default: ``True``).

		:returns: A `.Result` object.

		.. versionadded:: 2.0
		"""
		if not local:
			raise ValueError("Local path must not be empty!")
		
		is_file_like=hasattr(local,"write") and callable(local.write)
		
		# Massage remote path
		orig_remote=remote
		if is_file_like:
			local_base=getattr(local,"name",None)
		else:
			local_base=os.path.basename(local)
		if not remote:
			if is_file_like:
				raise ValueError(
					"Must give non-empty remote path when local is a file-like object!"  # noqa
					)
			else:
				remote=local_base
				# debug("Massaged empty remote path into {!r}".format(remote))
				pass
		elif self.is_remote_dir(remote):
			# non-empty local_base implies a) text file path or b) FLO which
			# had a non-empty .name attribute. huzzah!
			if local_base:
				remote=posixpath.join(remote,local_base)
			else:
				if is_file_like:
					raise ValueError(
						"Can't put a file-like-object into a directory unless it has a non-empty .name attribute!"  # noqa
						)
				else:
					# TODO: can we ever really end up here? implies we want to
					# reorganize all this logic so it has fewer potential holes
					raise ValueError(
						"Somehow got an empty local file basename ({!r}) when uploading to a directory ({!r})!".format(  # noqa
							local_base,remote
							)
						)
		
		prejoined_remote=remote
		remote=posixpath.join(
			self.sftp.getcwd() or self.sftp.normalize("."),remote
			)
		if remote!=prejoined_remote:
			# msg="Massaged relative remote path {!r} into {!r}"
			# debug(msg.format(prejoined_remote,remote))
			pass
		
		# Massage local path
		orig_local=local
		if not is_file_like:
			local=os.path.abspath(local)
			if local!=orig_local:
				# debug(
				# 	"Massaged relative local path {!r} into {!r}".format(
				# 		orig_local,local
				# 		)
				# 	)  # noqa
				pass
		
		# Run Paramiko-level .put() (side-effects only. womp.)
		# TODO: push some of the path handling into Paramiko; it should be
		# responsible for dealing with path cleaning etc.
		# TODO: probably preserve warning message from v1 when overwriting
		# existing files. Use logging for that obviously.
		#
		# If local appears to be a file-like object, use sftp.putfo, not put
		if is_file_like:
			# msg="Uploading file-like object {!r} to {!r}"
			# debug(msg.format(local,remote))
			
			pointer=local.tell()
			try:
				local.seek(0)
				self.sftp.putfo(fl=local,remotepath=remote,confirm=False)
			finally:
				local.seek(pointer)
		else:
			# debug("Uploading {!r} to {!r}".format(local,remote))
			self.sftp.put(localpath=local,remotepath=remote,confirm=False)
			# Set mode to same as local end
			# TODO: Push this down into SFTPClient sometime (requires backwards
			# incompat release.)
			if preserve_mode:
				local_mode=os.stat(local).st_mode
				mode=stat.S_IMODE(local_mode)
				self.sftp.chmod(remote,mode)
		# Return something useful
		return Result(
			orig_remote=orig_remote,
			remote=remote,
			orig_local=orig_local,
			local=local,
			connection=self.connection,
			)

class FabricConnection(Connection):
	
	def get(self,*args,**kwargs):
		return FabricTransfer(self).get(*args,**kwargs)
	
	def put(self,*args,**kwargs):
		"""
		Put a remote file (or file-like object) to the remote filesystem.

		Simply a wrapper for `.Transfer.put`. Please see its documentation for
		all details.

		.. versionadded:: 2.0
		"""
		return FabricTransfer(self).put(*args,**kwargs)

class SshMan(object):
	
	def __init__(self,fabric_connection: Connection,remote_workdir: lgblkb_tools.utils.Folder):
		self.conn: Connection=fabric_connection
		# if isinstance(remote_workdir,str): remote_workdir=remote_projects_dir.create(
		# 	remote_workdir)
		assert isinstance(remote_workdir,lgblkb_tools.utils.Folder),f"'remote_workdir' is neither a Folder nor a string."
		self.work_folder: lgblkb_tools.utils.Folder=remote_workdir
		# if self.conn.run(f'[ -d "{remote_workdir.path}" ] && echo "exist" || echo "not exist"',warn=True).stdout.strip()=='not exist':
		# 	logsup.logger.info('Creating folder %s on remote host.',remote_workdir.path)
		# 	self.conn.run(f'mkdir -p {remote_workdir.path}')
		pass
	
	def check_exists(self,program_name):
		which_result=self.conn.run(f'which {program_name}',warn=True)
		if which_result.failed: return False
		else:
			logsup.logger.info(f'"{program_name}" already exists.')
			return True
	
	def apt_install(self,*package_names,forced=False):
		names=list()
		for package_name in package_names:
			pack_names=[x for x in package_name.split(" ") if x]
			names.extend(pack_names)
		
		if not forced:
			for package_name in names.copy():
				if self.check_exists(package_name):
					names.remove(package_name)
					logsup.logger.info('Skipping installation of "%s".',package_name)
		
		for package_name in names:
			self.run(f'sudo apt-get install {package_name} -y')
			package_path=self.conn.run(f'which {package_name}',warn=True).stdout.strip()
			logsup.logger.debug('"%s" path: %s',package_name,package_path)
	
	@abstractmethod
	def initial_setup(self,*args,**kwargs):
		pass
	
	@abstractmethod
	def execute(self,*args,**kwargs):
		pass
	
	def run(self,command,**kwargs):
		logsup.logger.info('Running !!!%s!!!',command)
		return self.conn.run(command,**kwargs)
	
	def put(self,local,remote,preserve_mode=True,absolute_remote=False):
		if not absolute_remote: remote=self.work_folder.get_filepath(remote)
		logsup.logger.info('Putting from !!!%s!!! to !!!%s!!!',local,remote)
		out=self.conn.put(local,remote=remote,preserve_mode=preserve_mode)
		return out
	
	def get(self,remote,local=None,preserve_mode=True):
		logsup.logger.info('Getting from !!!%s!!! to !!!%s!!!',remote,local)
		out=self.conn.get(remote,local=local,preserve_mode=preserve_mode)
		return out
	
	def cd(self,path):
		return self.conn.cd(path=path)
	
	def cd_to_working_dir(self):
		return self.cd(self.work_folder.path)
	
	def prefix(self,command):
		return self.conn.prefix(command=command)
	
	def sudo(self,command,**kwargs):
		return self.conn.sudo(command=command,**kwargs)

class SshFolder(lgblkb_tools.utils.Folder):
	
	def _mkdir(self,child_dirpath):
		with self.sshman.cd_to_working_dir():
			self.sshman.conn.run(f'mkdir -p {child_dirpath}')
		return os.path.join(self.path,child_dirpath)
	
	def __init__(self,remote_path,pseudo=False,parent=None,**conn_kwargs):
		super(SshFolder,self).__init__(remote_path,pseudo=True,parent=parent,propagate_type=True)
		self.pseudo=pseudo
		if parent is None:
			self.fabric_conn=FabricConnection(**conn_kwargs)
		else:
			# self.sshman: SshMan=parent.sshman
			# self.sshman.work_folder=self
			self.fabric_conn=parent.fabric_conn
		self.sshman=SshMan(self.fabric_conn,self)
	
	def children(self):
		with self.sshman.cd_to_working_dir():
			item_names=self.sshman.run('ls').stdout.strip().split('\n')
		return list(map(lambda x:os.path.join(self.path,x),item_names))
	
	def delete(self):
		with self.sshman.cd_to_working_dir():
			self.sshman.run(f"""
cd ..
rm -r {self.path}
""")
	
	def clear(self):
		with self.sshman.cd_to_working_dir():
			self.sshman.run(f'rm -rf *')
	
	def md5(self,remote_filepath):
		return self.sshman.conn.run(f'md5sum {remote_filepath}',warn=True).stdout.strip().split(' ')[0]
	
	def upload_item(self,local_itempath,only_children=True):
		item_base_name=os.path.split(local_itempath)[-1]
		if os.path.isdir(local_itempath):
			if only_children:
				for item_to_upload in lgblkb_tools.utils.Folder(local_itempath).children():
					self.upload_item(item_to_upload,only_children=False)
				pass
			else:
				remote_folder=self.create(item_base_name)
				for local_sub_item_path in lgblkb_tools.utils.Folder(local_itempath).children():
					remote_folder.upload_item(local_sub_item_path)
		else:
			local_checksum=gsup.md5(local_itempath)
			remote_filepath=self.get_filepath(item_base_name)
			while local_checksum!=self.md5(remote_filepath):
				self.sshman.run(f"""rm {remote_filepath}""",warn=True)
				self.sshman.put(local_itempath,remote_filepath,absolute_remote=True)
	
	def run(self,command,**kwargs):
		return self.sshman.run(command=command,**kwargs).stdout.strip()
	
	def run_with_cd(self,command,**kwargs):
		with self.sshman.cd_to_working_dir():
			return self.run(command,**kwargs)
	
	def get(self,remote,local=None,preserve_mode=True):
		return self.sshman.get(remote,local=local,preserve_mode=preserve_mode)
	
	def put(self,local,remote,preserve_mode=True,absolute_remote=False):
		return self.sshman.put(local,remote,preserve_mode=preserve_mode,absolute_remote=absolute_remote)

def main():
	pass

if __name__=='__main__':
	main()



from typing import Union
import sys
from timeit import default_timer as timer
import logging
from logging.handlers import TimedRotatingFileHandler as TimedHandler
import logging.handlers as loghandlers
import functools
from python_log_indenter import IndentedLoggerAdapter
from datetime import datetime
from .locations import get_name,create_path,InfoDict
from colorlog import ColoredFormatter

def create_filter(filter_func,*args,filter_name='',**kwargs):
	class CustomFilter(logging.Filter):
		
		def filter(self,record):
			return filter_func(record.getMessage(),*args,**kwargs)
	
	return CustomFilter(filter_name)

logging.INFORM=INFORM=369
logging.addLevelName(INFORM,"INFORM")

def inform(self,message,*args,**kws):
	# Yes, logger takes its '*args' as 'args'.
	if self.isEnabledFor(INFORM):
		self._log(INFORM,message,args,**kws)

logging.Logger.inform=inform

# region level_mapper:
level_mapper=dict()
level_mapper[logging.DEBUG]=lambda some_logger:some_logger.debug
level_mapper[logging.INFO]=lambda some_logger:some_logger.info
level_mapper[logging.WARNING]=lambda some_logger:some_logger.warning
level_mapper[logging.ERROR]=lambda some_logger:some_logger.error
level_mapper[logging.CRITICAL]=lambda some_logger:some_logger.critical
level_mapper[logging.INFORM]=lambda some_logger:some_logger.inform

# endregion
#todo:Group log files into folder. After expiration, delete folder (if left empty after cleaning).


class TheLogger(IndentedLoggerAdapter):
	
	def __init__(self,name,folder_path='',level=logging.DEBUG,log_format: str = None,to_stream=True,**kwargs):
		self.base_log_folder: Folder=Folder(folder_path) if folder_path else None
		super(TheLogger,self).__init__(logging.Logger(name,level),**dict(dict(spaces=1,indent_char='|--'),**kwargs))
		
		# IndentedLoggerAdapter.__init__(self,logging.Logger(name,level),**dict(dict(spaces=1,indent_char='|---')))
		# Folder.__init__(self)
		
		# super(IndentedLoggerAdapter,self).__init__(logging.Logger(name,level),**dict(dict(spaces=1,indent_char='|---')))
		# super(Folder,self).__init__()
		# super(TheLogger,self).__init__(**kwargs))
		# Folder.__init__(TheLogger,self.base_log_folder.path)
		# self.formatter=logging.Formatter(log_format or color_log_format)
		# IndentedLoggerAdapter.__init__(self,logging.Logger(name,level),**dict(dict(spaces=1,indent_char='|---'),**kwargs))
		# super(Folder,self).__init__()
		
		_log_format=(log_format or nocolor_log_format).replace(color_info,'')
		self.formatter=logging.Formatter(_log_format)
		if to_stream:
			stream_format=(log_format or color_log_format)  #.replace(color_info,'')
			stream_formatter=ColoredFormatter(stream_format)
			stream_handler=logging.StreamHandler()
			# stream_handler.setLevel(level)
			# stream_handler.setFormatter(logging.Formatter(stream_format))
			self.add_handler(stream_handler,formatter=stream_formatter)
		self.current_log_folder: Folder=self.base_log_folder
		self.log_path=''
	
	def add_handler(self,logHandler=None,level=None,log_format=None,filepath=None,formatter=None):
		if logHandler is None:
			filepath=filepath or self.log_path
			if not filepath:
				raise KeyError("Provide the filepath or log_path.")
			logHandler=logging.FileHandler(filepath)
			self.debug('log_filepath=%s',filepath)
		logHandler.setLevel(level or self.logger.level)
		
		if formatter is None:
			if log_format is None: formatter=self.formatter
			else: formatter=logging.Formatter(log_format)
		logHandler.setFormatter(formatter)
		self.logger.addHandler(logHandler)
		self.pop()
		return logHandler
	
	def add_timed_handler(self,when='d',interval=1,backupCount=14,level=None,log_format=None,filepath=None,**other_opts):
		filepath=filepath or self.log_path
		if filepath is None: raise KeyError("Provide the filepath or log_path.")
		log_handler=self.add_handler(loghandlers.TimedRotatingFileHandler(
			filename=filepath,when=when,interval=interval,backupCount=backupCount,**other_opts),
			level=level,log_format=log_format)
		return log_handler
	
	def add_rotating_handler(self,maxBytes=2e6,backupCount=14,level=None,log_format=None,filepath=None,**other_opts):
		filepath=filepath or self.log_path
		if filepath is None: raise KeyError("Provide the filepath or log_path.")
		log_handler=self.add_handler(loghandlers.RotatingFileHandler(
			maxBytes=int(maxBytes),filename=filepath,backupCount=backupCount,**other_opts),level=level,log_format=log_format)
		return log_handler
	
	def create(self,*child_folders,**info_kwargs):
		self.current_log_folder=self.base_log_folder.create(*[get_name(x) for x in child_folders],**info_kwargs)
		return self
	
	def get_filepath(self,*name_portions,ext='.log',delim='__',include_depth=2,include_datetime=True,**name_kwargs):
		datetime_loc_index=None
		if type(include_datetime) is bool:
			if include_datetime: datetime_loc_index=include_depth
		else: datetime_loc_index=include_datetime
		
		# self.info('datetime_loc_index: %s',datetime_loc_index)
		
		self.log_path=self.current_log_folder.get_filepath(*name_portions,ext=ext,delim=delim,include_depth=include_depth,
		                                                   datetime_loc_index=datetime_loc_index,**name_kwargs)
		return self
	
	# def create_logpath(self,*name_portions,ext='',delim='_',include_depth=None,datetime_loc_index=None,**name_kwargs):
	# 	fp=self.current_log_folder.get_filepath(*name_portions,ext=ext,delim=delim,include_depth=include_depth,
	# 	                                       datetime_loc_index=datetime_loc_index,**name_kwargs)
	# 	self.add_timed_handler(fp,**(timing_opts or {}))
	# 	# self.info('log_filepath=%s',fp)
	# 	self.inform('log_filepath=%s',fp)
	# 	self.pop()
	# 	return self
	
	# def create_log_file(self,filename,include_depth=2,timing_opts=None,folder_parts=None,file_parts=None,**kwargs):
	# 	fp=self.base_log_folder.create(file=get_name(filename),**(folder_parts or {}))\
	# 		.get_filepath(include_depth=include_depth,include_datetime=-include_depth,**(file_parts or {}),**kwargs,pid=os.getpid())
	# 	self.add_timed_handler(fp,**(timing_opts or {}))
	# 	# self.info('log_filepath=%s',fp)
	# 	self.inform('log_filepath=%s',fp)
	# 	self.pop()
	# 	return self
	
	def __getitem__(self,item):
		return level_mapper[item](self)
	
	def inform(self,msg,*args,**kwargs):
		self.logger.inform(msg,*args,**kwargs)
	
	def with_logging(self,log_level=logging.DEBUG,atomic_print=False,show_inputs=False):
		# if logger is None: logger=simple_logger
		# assert logger is not None
		logger_say=level_mapper[log_level](self)
		
		def second_wrapper(func):
			@functools.wraps(func)
			def wrapper(*args,**kwargs):
				if not atomic_print:
					if show_inputs:
						logger_say('Running "%s" with args=%s and kwargs=%s:',func.__name__,args,kwargs)
					else:
						logger_say('Running "%s":',func.__name__)
				self.add()
				
				start=timer()
				try:
					result=func(*args,**kwargs)
				except KeyboardInterrupt:
					logger_say('KeyboardInterrupt within %s. Duration: %s',
					           func.__name__,timer()-start)
					sys.exit()
				except Exception as e:
					self.error(str(e),exc_info=True)
					raise e
				self.sub()
				logger_say('Done "%s". Duration: %.3f sec.',func.__name__,timer()-start)
				return result
			
			return wrapper
		
		return second_wrapper

log_fmt="%(asctime)s -- %(levelname)s -- %(name)s -- %(funcName)s -- %(filename)s -- %(lineno)d -- %(message)s"
# color_log_format="[%(asctime)s] [pid:%(process)5s] [%(levelname)8s]: %(message)s"
# color_log_format="[%(asctime)s] [%(levelname)-8s%(reset)s]: %(message)s"
color_info='%(log_color)s'
aligner=lambda n,_type='s':f'{n}{_type}'
color_log_format=f"{color_info}%(asctime)s pid:%(process){aligner(5,'d')} %(levelname){aligner(8)} {color_info}%(message)s"
# nocolor_log_format=f"%(asctime)s %(levelname){aligner(8)} %(message)s"
nocolor_log_format=f"%(asctime)s pid:%(process){aligner(5,'d')} %(levelname){aligner(8)} %(message)s"
# simple_fmt_without_color=f"%(asctime)s %(levelname){aligner(8)} %(message)s"
simple_fmt_no_level="%(asctime)s|||: %(message)s"
#log_folder=None  #Folder('~').create('backend_logs')
logger=TheLogger('simple_logger')

def with_logging(log_level=logging.DEBUG,atomic_print=False,show_inputs=False):
	# if logger is None: logger=simple_logger
	# assert logger is not None
	def second_wrapper(func):
		@functools.wraps(func)
		def wrapper(*args,**kwargs):
			logger_say=level_mapper[log_level](logger)
			if not atomic_print:
				if show_inputs:
					logger_say('Running "%s" with args=%s and kwargs=%s:',func.__name__,args,kwargs)
				else:
					logger_say('Running "%s":',func.__name__)
			logger.add()
			
			start=timer()
			try:
				result=func(*args,**kwargs)
			except KeyboardInterrupt:
				logger_say('KeyboardInterrupt within %s. Duration: %s',
				           func.__name__,timer()-start)
				sys.exit()
			except Exception as e:
				logger.error(str(e),exc_info=True)
				raise e
			finally:
				logger.sub()
			logger_say('Done "%s". Duration: %.3f sec.',func.__name__,timer()-start)
			return result
		
		return wrapper
	
	return second_wrapper

# def with_logging(log_level=logging.INFO,atomic_print=False,show_inputs=False):
# 	# global log_folder,logger
# 	# if log_folder is None: log_folder=Folder('~').create('backend_logs')
# 	# if logger is None: logger=TheLogger('simple_logger',log_folder)
# 	assert logger is not None,"Please, first create logger using create_logger(folder_path) method."
# 	return logger.with_logging(log_level,atomic_print=atomic_print,show_inputs=show_inputs)



# def get_logger_filepath(info: dict,dir_depth=1):
# 	kv_pairs=[f'{k}={v}' for k,v in info.items()]
# 	log_filename="___".join([*kv_pairs[dir_depth:],*kv_pairs[:dir_depth]])
# 	log_filepath=create_path(1,logs_folder.path,*kv_pairs[:dir_depth],log_filename+'.log')
# 	return log_filepath2

# simple_logger=TheLogger('simple_logger',)  #create_process_logger(__file__,collections.OrderedDict(pid=os.getpid()))

def create_logger(folder_path,name='default_logger',level=logging.DEBUG,**kwargs):
	from .utils import Folder
	logger.base_log_folder=logger.current_log_folder=Folder(folder_path)
	logger.logger.name=name
	logger.logger.level=level
	# global logger
	# logger=TheLogger(name,folder_path,level=level,**kwargs)
	# log_folder=logger.base_log_folder
	return logger

def main():
	return

if __name__=='__main__':
	main()
	pass



import json

import geojson
import geojsonio
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj
import shapely.geometry as shg
import shapely.ops as shops
import shapely.wkb as shwkb
import shapely.wkt as shwkt
from functools import partial
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon

geom_kz=r'POLYGON ((59.501953125 51.39920565355378, 51.50390625 52.50953477032727, 46.40625 50.875311142200765, 45.1318359375 48.516604348867475, 51.37207031249999 41.04621681452063, 56.4697265625 40.713955826286046, 58.00781249999999 44.49650533109348, 68.466796875 40.04443758460856, 80.595703125 41.60722821271717, 85.869140625 46.46813299215554, 89.1650390625 49.009050809382046, 77.2998046875 54.87660665410869, 70.09277343749999 56.07203547180089, 60.1171875 54.316523240258256, 59.501953125 51.39920565355378))'

def remove_redundants(polygons,redundancy_order=2.5,**kwargs):
	area_max=max([g.area for g in polygons])
	candidates=[g for g in polygons]
	for g in list(polygons):
		if np.log10(area_max/g.area)>redundancy_order:
			print('Redundant land is detected. Removing it...')
			candidates.remove(g)
	return candidates

def get_crs(geojson_datum):
	crs_info=geojson_datum.crs.properties['name'].split(':')
	if not crs_info[-1].isnumeric():
		print(crs_info)
		raise NotImplementedError('Could not infer crs. crs_info = \n{}'.format(crs_info))
	return int(crs_info[-1])

def plot_polygon(polygon: shg.Polygon,label=None,ax=None,**kwargs):
	if ax is None: plot=plt.plot
	else: plot=ax.plot
	default_opts=dict(lw=3)
	if label is None: labeler=lambda l:dict(default_opts,label=l)
	else: labeler=lambda l:dict(default_opts,label='{} {}'.format(label,l))
	plotter=lambda linestring,_label:plot(*linestring.xy,**dict(labeler(_label),**kwargs))
	plotter(polygon.exterior,'exterior')
	for i,p_i in enumerate(polygon.interiors):
		plotter(p_i,'interior {}'.format(i))

def plot_poly(p,**kwargs):
	if isinstance(p,shg.MultiPolygon):
		for p in p: plot_poly(p,**kwargs)
	plt.plot(*p.exterior.xy,**kwargs)

def convert_crs(coor_obj,_in: int,_out=4326):
	inproj,outproj=pyproj.Proj(init=f'epsg:{_in}'),pyproj.Proj(init=f'epsg:{_out}')
	converter=lambda xy:np.stack(pyproj.transform(inproj,outproj,*xy),axis=1).reshape(-1,2)
	polygons=list()
	if type(coor_obj) is np.ndarray:
		assert coor_obj.shape[1]==2,\
			"ndarray shape should be (-1,2), but shape={} is provided.".format(coor_obj.shape)
		return converter(np.split(coor_obj,2,axis=1))
	elif type(coor_obj) is shg.Polygon:
		coor_obj=shg.MultiPolygon([coor_obj])
		returner=lambda multipolygon:multipolygon[0]
	elif type(coor_obj) is shg.MultiPolygon:
		returner=lambda multipolygon:multipolygon
		pass
	else: raise NotImplementedError('No implementation for objects of type = {}'.format(type(coor_obj)))
	for g in coor_obj:
		interior_xys=[converter(interior.xy) for interior in g.interiors]
		polygons.append(shg.Polygon(converter(g.exterior.xy),interior_xys))
	return returner(shg.MultiPolygon(polygons))

def reproject_point(point,_in,_out):
	inproj,outproj=pyproj.Proj(init=f'epsg:{_in}'),pyproj.Proj(init=f'epsg:{_out}')
	return shg.Point(pyproj.transform(inproj,outproj,point.x,point.y))

def get_area(geom,in_epsg):
	geom_area=shops.transform(
		partial(
			pyproj.transform,
			pyproj.Proj(init=f'EPSG:{in_epsg}'),
			pyproj.Proj(proj='aea',lat1=geom.bounds[1],lat2=geom.bounds[3])),geom).area
	return geom_area

class SpatialGeom:
	
	@classmethod
	def from_geojson(cls,geojson_datum,crs_in=None,crs_out=4326,**kwargs):
		if isinstance(geojson_datum,str):
			geojson_datum=geojson.load(open(geojson_datum))
		if 'features' in geojson_datum.keys(): geometry=geojson_datum.features[0].geometry
		else: geometry=geojson_datum.geometry
		cad_land=SpatialGeom(shg.shape(geometry),**kwargs)
		if cad_land.geom_obj.is_empty:
			print('Empty entry detected.')
			return cad_land
		cad_land.data['epsg']=crs_out
		if not 'name' in kwargs: cad_land.name=geojson_datum['features'][0]['properties']['KAD_NOMER']
		if crs_in==crs_out: return cad_land
		elif crs_in is None:
			if 'crs' in geojson_datum.keys():
				crs_in=get_crs(geojson_datum)
				if crs_in==crs_out: return cad_land
			elif max(cad_land.geom_obj.bounds)>90:
				#print('Assuming input crs = epsg: 32643')
				#crs_in='epsg:32643'
				raise NotImplementedError(
					'No info to infer from. Please, specify incoming coordinate reference system.')
			else:
				#print('Assuming proper input crs, i.e. "epsg:4326".')
				#return cad_land
				raise NotImplementedError(
					'No info to infer from. Please, specify incoming coordinate reference system.')
		cad_land.convert_crs(crs_in,crs_out)
		return cad_land
	
	def convert_crs(self,_from: int,to=4326):
		if self.geom_obj.is_empty:
			print('Cannot convert empty shape.')
			return self
		if _from==to: return self
		multi_polygon=convert_crs(self.geom_obj,_from,to)
		#if inplace:
		self.geom_obj=multi_polygon
		return self
		# else:
		# 	return SpatialGeom(multi_polygon,redundancy_order=self.redundancy_order,
		# 	                   name=self.name)
		pass
	
	def __init__(self,geom_obj,name=None,**kwargs):
		"""

		:param geom_obj: wkt/wkb string, ndarray or shapely Polygon/Multipolygon
		:param name: name
		:param kwargs:
		redundancy_order = 2.5
		"""
		if isinstance(geom_obj,str):
			try:
				geom_obj=shwkt.loads(geom_obj)
			except Exception as e:
				try:
					json_obj=geojson.loads(geom_obj)
					geom_obj=shg.shape(json_obj['features'][0]['geometry'])
				except Exception as e:
					from shapely import geos
					# geos.WKBWriter.defaults['include_srid']=True
					geom_obj=shwkb.loads(geom_obj,hex=True)
		
		if isinstance(geom_obj,shg.Polygon): geom_obj=shg.MultiPolygon([geom_obj])
		elif isinstance(geom_obj,np.ndarray):
			geom_obj=shg.MultiPolygon([shg.Polygon(geom_obj)])
		self.name=name or 'SpatialGeom'
		self.data=kwargs
		if not geom_obj.is_empty: geom_obj=shg.MultiPolygon(remove_redundants(geom_obj,**kwargs))
		self.geom_obj: shg.MultiPolygon=geom_obj
		self.data['with_holes']=max(self.do_for_each(lambda p:len(list(p.interiors))),default=0)!=0
		pass
	
	def do_for_each(self,func,*args,**kwargs):
		doer=lambda x:func(x,*args,**kwargs)
		#if self.is_multipolygon:
		out=list()
		for p in self.geom_obj:
			out.append(doer(p))
		return out
	
	# else:
	# 	return [doer(self.geom_obj)]
	
	def show_on_geojsonio(self):
		get_displayed=lambda obj:geojsonio.display(json.dumps(shg.mapping(obj)))
		self.do_for_each(get_displayed)
	
	def plot(self,show=False,**kwargs):
		if self.geom_obj.is_empty: return
		# if ax is None: plot=plt.plot
		# else: plot=ax.plot
		# plotter=lambda linearring,label,:plot(*linearring.xy,**dict(dict(label=label),**kwargs))
		#if self.is_multipolygon:
		for i,p in enumerate(self.geom_obj):
			plot_polygon(p,**dict(dict(label='Polygon {}'.format(i)),**kwargs))
		if show:
			plt.legend(loc='best')
			plt.show()
	
	def to_geojson(self,**kwargs):
		return geojson.Feature(geometry=self.geom_obj,**kwargs)
	
	def __bool__(self):
		return self.geom_obj.is_empty
	
	# else:
	# 	plot_polygon(self.geom_obj,'Single polygon',**kwargs)
	
	def get_area(self):
		return get_area(self.geom_obj,4326)
	
	@property
	def box(self):
		return self.geom_obj.envelope.exterior
	
	@property
	def xy(self):
		return np.array(self.geom_obj[0].boundary.xy).T
	
	def __str__(self):
		return '{}: {}'.format(self.name,str(self.geom_obj))

def plot_patches(polygons,ax=None,c=None,**kwargs):
	if ax is None: ax=plt.gca()
	if isinstance(polygons,pd.Series):
		pcol=PatchCollection(polygons.map(lambda p:Polygon(p.exterior)),**kwargs)
	else:
		poly_objs=[Polygon(x.exterior) for x in polygons]
		pcol=PatchCollection(poly_objs,**kwargs)
	if c: pcol.set_color(c)
	ax.add_collection(pcol)

def get_rect(corner,sizes):
	corner=np.array(corner)
	sizes=np.array(sizes)
	return np.array([corner,corner+sizes*(1,0),corner+sizes,corner+sizes*(0,1)])

def get_portion(polygon,offset_ratio,size_ratio):
	p_min=np.array(polygon.bounds[0:2])
	p_max=np.array(polygon.bounds[-2:])
	curr_sizes=(p_max-p_min)
	new_rect=shg.Polygon(get_rect(p_min+curr_sizes*offset_ratio,curr_sizes*size_ratio))
	return new_rect

def get_forecast_region(cad_geom,min_forecast_point_dist=3,forecast_step_degrees=0.25):
	forecast_region=cad_geom.envelope.buffer(min_forecast_point_dist*forecast_step_degrees*1.2)
	return forecast_region.envelope

if __name__=='__main__':
	p=shg.Point(78,43)
	print(reproject_point(p,4326,3857))
	# print(p.x)
	pass

# a1=shg.Polygon(np.random.rand(4,2))
# a2=shg.Polygon()
# b1=shg.MultiPolygon()
# b2=shg.MultiPolygon([a1])
# b3=shg.MultiPolygon([a1,shg.Polygon(np.random.rand(4,2),[np.random.rand(3,2)])])
# c1=SpatialGeom(a1)
# #c2=c1.convert_crs(32642)
# print(convert_crs(np.random.rand(10,2),32642))
def show_images(images,cols=1,titles=None,show=True,interp='none'):
	"""Display a list of images in a single figure with matplotlib.

	Parameters
	---------
	images: List of np.arrays compatible with plt.imshow.

	cols (Default = 1): Number of columns in figure (number of rows is
						set to np.ceil(n_images/float(cols))).

	titles: List of titles corresponding to each image. Must have
			the same length as titles.
	"""
	assert ((titles is None) or (len(images)==len(titles)))
	n_images=len(images)
	if titles is None: titles=['Image (%d)'%i for i in range(1,n_images+1)]
	fig=plt.figure()
	for n,(image,title) in enumerate(zip(images,titles)):
		a=fig.add_subplot(cols,np.ceil(n_images/float(cols)),n+1)
		if image.ndim==2: plt.gray()
		plt.imshow(image,interpolation=interp)
		a.set_title(title)
	fig.set_size_inches(np.array(fig.get_size_inches())*n_images)
	if show: plt.show()



import collections
import fnmatch
import os
import shutil
import zipfile

get_name=lambda some_path:os.path.splitext(os.path.basename(some_path))[0]

def get_parent_dir(some_path,depth=1):
	for i in range(depth):
		some_path=os.path.dirname(os.path.abspath(some_path))
	return some_path

def create_path(*paths,stop_depth=0):
	path=os.path.join(*paths)
	os.makedirs(get_parent_dir(path,stop_depth),exist_ok=True)
	return path

def get_existing_path(paths,silent=False):
	for p in paths:
		if os.path.exists(p): return p
	if not silent:
		raise NotImplementedError('Could not find any existing path.')

def clear_folder(folder,remove_subdirs=True):
	for the_file in os.listdir(folder):
		file_path=os.path.join(folder,the_file)
		try:
			if os.path.isfile(file_path):
				os.unlink(file_path)
			elif remove_subdirs and os.path.isdir(file_path): shutil.rmtree(file_path)
		except Exception as e:
			print(e)

def get_full_paths_from(parent_dir: str):
	for file in os.listdir(parent_dir):
		yield os.path.join(parent_dir,file)

def get_splitted(path):
	folders=[]
	while 1:
		path,folder=os.path.split(path)
		
		if folder!="":
			folders.append(folder)
		else:
			if path!="":
				folders.append(path)
			break
	folders.reverse()
	return folders

class InfoDict(collections.OrderedDict):
	
	def get_dir_path(self,*paths):
		return os.path.join(*self.get_portions(),*paths)
	
	# def create_file_path(self,ext,dir_depth=1):
	# 	return create_info_path(self,ext=ext,*self.get_portions(),dir_depth=dir_depth)
	
	def get_portions(self):
		portions=list()
		for k,v in self.items():
			if v: portions.append(f"{k}_{v}")
			else: portions.append(k)
		return portions
	
	def get_text(self,delim="__"):
		return delim.join(self.get_portions())
	
	def __str__(self):
		return self.get_text()

class ZipError(Exception):
	pass

class CopyError(Exception):
	pass

def _make_zipfile(base_name,base_dir,root_dir=None,dry_run=0,logger=None):
	"""Create a zip file from all the files under 'base_dir'.

	The output zip file will be named 'base_name' + ".zip".  Returns the
	name of the output zip file.
	"""
	save_cwd=os.getcwd()
	if root_dir: os.chdir(root_dir)
	zip_filename=base_name+".zip"
	archive_dir=os.path.dirname(base_name)
	
	if archive_dir and not os.path.exists(archive_dir):
		if logger is not None:
			logger.debug("creating %s",archive_dir)
		if not dry_run:
			os.makedirs(archive_dir)
	
	if logger is not None:
		logger.debug("creating '%s' and adding '%s' to it",
		             zip_filename,base_dir)
	
	if not dry_run:
		with zipfile.ZipFile(zip_filename,"w",
		                     compression=zipfile.ZIP_DEFLATED) as zf:
			path=os.path.normpath(base_dir)
			if path!=os.curdir:
				zf.write(path,path)
				if logger is not None:
					logger.debug("adding '%s'",path)
			for dirpath,dirnames,filenames in os.walk(base_dir):
				for name in sorted(dirnames):
					path=os.path.normpath(os.path.join(dirpath,name))
					zf.write(path,path)
					if logger is not None:
						logger.debug("adding '%s'",path)
				for name in filenames:
					path=os.path.normpath(os.path.join(dirpath,name))
					if os.path.isfile(path):
						zf.write(path,path)
						if logger is not None:
							logger.debug("adding '%s'",path)
	os.chdir(save_cwd)
	return zip_filename

# class RemoteFolder(Folder):
#
# 	def __init__(self,path=''):
# 		super().__init__(path='')
# 		self.path=path
#
# 	def create(self,*child_folders,**info_kwargs):
# 		return super().create(*child_folders,is_remote=True,**info_kwargs)
#
# 	def delete(self):
# 		raise NotImplementedError
#
# 	def children(self,*paths):
# 		raise NotImplementedError

get_neighbor_path=lambda curr_path,filename:os.path.join(get_parent_dir(curr_path),filename)

def get_paths_matching(target_dir,match: str):
	items=os.listdir(target_dir)
	outs=list()
	for item in items:
		if match in item: outs.append(os.path.join(target_dir,item))
	return outs

def replace_ext(fpath,out_extension):
	base_path,ext=os.path.splitext(fpath)
	return base_path+out_extension

def find_file(directory,pattern):
	for root,dirs,files in os.walk(directory):
		for basename in files:
			if fnmatch.fnmatch(basename,pattern):
				filename=os.path.join(root,basename)
				return filename

def main():
	# project_folder=Folder(__file__)
	# logs_folder=project_folder.create('log_files')
	# results_folder=project_folder.create('Results')
	# test_folder=Folder(r'/home/lgblkb/PycharmProjects/Egistic/Scripts/lgblkb_local_folder_2')
	# dst_folder=Folder(r'/home/lgblkb/PycharmProjects/Egistic/lgblkb_scripts')
	# test_folder.copy_to(dst_folder)
	
	pass

if __name__=='__main__':
	main()



class MainStatus:
	delimiter='_'
	
	def __init__(self,prefix,is_fallback):
		self.prefix=prefix
		self.is_fallback=is_fallback
	
	def __get_status(self,status_text):
		out_status=MainStatus.delimiter.join([self.prefix,status_text])
		if self.is_fallback: return fbsm[out_status]
		else: return out_status
	
	@property
	def Download(self): return self.__get_status('Download')
	
	@property
	def HistoricDownload(self): return self.__get_status('HistoricDownload')
	
	@property
	def ImageCheck(self): return self.__get_status('ImageCheck')
	
	@property
	def AtmosphericCorrection(self): return self.__get_status('AtmosphericCorrection')
	
	@property
	def IndexProcessing(self): return self.__get_status('IndexProcessing')

class Status:
	__unprocessed='Unprocessed'
	
	def __init__(self,is_fallback=False):
		self.__is_fallback=is_fallback
	
	def __get_normal_status(self,out_status):
		if self.__is_fallback: return fbsm[out_status]
		else: return out_status
	
	@property
	def Unprocessed(self): return self.__get_normal_status(self.__unprocessed)
	
	unprocessed_vals=[__unprocessed,'0','',None]
	FinishedRequest='FinishedRequest'
	SearchingForCloudFreeProducts='SearchingForCloudFreeProducts'
	Error_NoOptimalProduct='Error_NoOptimalProduct'
	Error_Check_AtmosCor_ProductDoesNotExist='Error_Check_AtmosCor_ProductDoesNotExist'
	Error_Sen2Cor='Error_Sen2Cor'
	Error_DownloadResultsEmpty='Error_DownloadResultsEmpty'
	NoAvailableProducts='NoAvailableProducts'
	NoAvailableHistoricData='NoAvailableHistoricData'
	NotForProcessing='NotForProcessing'
	
	def __get_complex_status(self,prefix):
		return MainStatus(prefix,self.__is_fallback)
	
	@property
	def QueuedFor(self): return self.__get_complex_status('QueuedFor')
	
	@property
	def Running(self): return self.__get_complex_status("Running")
	
	@property
	def Finished(self): return self.__get_complex_status("Finished")
	
	@property
	def SubmittedTo(self): return self.__get_complex_status("SubmittedTo")
	
	@property
	def WaitingFinish(self): return self.__get_complex_status("WaitingFinish")
	
	@property
	def fallback_to(self):
		return Status(is_fallback=True)

image_status=Status()
fallback_status_mapper=dict()
fallback_status_mapper[image_status.Unprocessed]=['0','',
                                                  image_status.Running.Download,
                                                  image_status.QueuedFor.Download,
                                                  image_status.QueuedFor.HistoricDownload,
                                                  image_status.Running.ImageCheck,
                                                  image_status.Error_DownloadResultsEmpty,
                                                  image_status.Error_NoOptimalProduct,
                                                  image_status.NoAvailableProducts]
fallback_status_mapper[image_status.SubmittedTo.AtmosphericCorrection]=[image_status.Running.AtmosphericCorrection,
                                                                        image_status.QueuedFor.AtmosphericCorrection,
                                                                        image_status.WaitingFinish.AtmosphericCorrection]
fallback_status_mapper[image_status.SubmittedTo.IndexProcessing]=[image_status.QueuedFor.IndexProcessing,
                                                                  image_status.Running.IndexProcessing]
fallback_status_mapper[image_status.SubmittedTo.ImageCheck]=[image_status.QueuedFor.ImageCheck,
                                                             image_status.QueuedFor.Download,
                                                             image_status.Running.ImageCheck,
                                                             image_status.Running.Download,
                                                             image_status.NoAvailableProducts]
fbsm=fallback_status_mapper

def main():
	pass

if __name__=='__main__':
	main()
